import smtplib
from email.mime.application import MIMEApplication
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from typing import Union
import imaplib


def send(
    sender: str,
    password: str,
    receivers: Union[str, list[str]],
    subject: str,
    content: str,
    pngfiles: list[str] = (),
):
    msg = MIMEMultipart()
    msg["Subject"] = subject
    msg["From"] = sender
    msg["To"] = ", ".join(receivers if receivers is list else [receivers])
    msg.attach(MIMEText(content, "plain"))
    for file in pngfiles:
        extension = file.split(r".")[-1]
        with open(file, "rb") as fp:
            if extension == "png":
                attached_file = MIMEImage(fp.read())
                # Open the files in binary mode.  Let the MIMEImage class automatically
                # guess the specific image type.
            else:
                attached_file = MIMEApplication(fp.read(), Name=file)
        msg.attach(attached_file)

    send_port = 587

    # Send the email via our own SMTP server.
    host = "smtp.gmail.com" if "gmail" in sender else "smtp.office365.com"
    with smtplib.SMTP(host, send_port) as smtp:
        smtp.ehlo()
        smtp.starttls()
        smtp.login(sender, password)
        smtp.sendmail(sender, receivers, msg.as_string())
        smtp.quit()


def receive(receiver: str, password: str):
    # connect to the server and go to its inbox
    mail = imaplib.IMAP4_SSL("imap.gmail.com")
    mail.login(receiver, password)
    # we choose the inbox but you can select others
    mail.select("inbox")
    # we'll search using the ALL criteria to retrieve
    # every message inside the inbox
    # it will return with its status and a list of ids
    status, data = mail.search(None, "ALL")
    # the list returned is a list of bytes separated
    # by white spaces on this format: [b'1 2 3', b'4 5 6']
    # so, to separate it first we create an empty list
    mail_ids = []
    # then we go through the list splitting its blocks
    # of bytes and appending to the mail_ids list
    for block in data:
        # the split function called without parameter
        # transforms the text or bytes into a list using
        # as separator the white spaces:
        # b'1 2 3'.split() => [b'1', b'2', b'3']
        mail_ids += block.split()

    # now for every id we'll fetch the email
    # to extract its content
    for i in mail_ids:
        # the fetch function fetch the email given its id
        # and format that you want the message to be
        status, data = mail.fetch(i, "(RFC822)")

        # the content data at the '(RFC822)' format comes on
        # a list with a tuple with header, content, and the closing
        # byte b')'
        for response_part in data:
            # so if its a tuple...
            if isinstance(response_part, tuple):
                # we go for the content at its second element
                # skipping the header at the first and the closing
                # at the third
                message = email.message_from_bytes(response_part[1])

                # with the content we can extract the info about
                # who sent the message and its subject
                mail_from = message["from"]
                mail_subject = message["subject"]

                # then for the text we have a little more work to do
                # because it can be in plain text or multipart
                # if its not plain text we need to separate the message
                # from its annexes to get the text
                if message.is_multipart():
                    mail_content = ""

                    # on multipart we have the text message and
                    # another things like annex, and html version
                    # of the message, in that case we loop through
                    # the email payload
                    for part in message.get_payload():
                        # if the content type is text/plain
                        # we extract it
                        if part.get_content_type() == "text/plain":
                            mail_content += part.get_payload()
                else:
                    # if the message isn't multipart, just extract it
                    mail_content = message.get_payload()

                # and then let's show its result
                print(f"From: {mail_from}")
                print(f"Subject: {mail_subject}")
                print(f"Content: {mail_content}")



import re
import os
import glob
from typing import Literal, Tuple
import pyarrow.parquet as pq
import xlrd
import pandas as pd
import numpy as np
import pyarrow as pa
from scipy.interpolate import NearestNDInterpolator
from colorama import Fore, Back, Style


def caster(df: pd.DataFrame):
    string_cols = [col for col, col_type in df.dtypes.items() if col_type == "object"]
    if len(string_cols) > 0:
        mask = df.astype(str).apply(
            lambda x: x.str.match(r"(\d{1,4}[-/\\\. ]\d{1,2}[-/\\\. ]\d{2,4})+.*").any()
        )

        def excel_date(x):
            # Warning : Date in the French format
            x = pd.to_datetime(x, dayfirst=True, errors="ignore")
            x = x.apply(
                lambda y: xlrd.xldate.xldate_as_datetime(y, 0)
                if type(y) in (int, float)
                else y
            )
            return x.astype("datetime64[ns]")

        df.loc[:, mask] = df.loc[:, mask].apply(excel_date)
        del mask
    string_cols = [col for col, col_type in df.dtypes.items() if col_type == "object"]
    for col in string_cols:
        try:
            try:
                df[col] = df[col].str.split(",").str.join(".").values.astype(float)
            except AttributeError:
                df[col] = df[col].values.astype(float)
        except (ValueError, TypeError):
            pass

    df = df.apply(
        lambda col: col
        if not (
            col.dtype in [np.dtype("float64"), np.dtype("float32"), np.dtype("float16")]
        )
        or col.isna().any()
        or pd.Series((col.apply(np.int64) == col)).sum() != df.shape[0]
        else col.apply(np.int64),
        axis=0,
    )
    return df


def load(
    file: str,
    sheet: int = 1,
    skiprows=None,
    on: str = "index",
    classic_data=True,
    recursive=True,
):
    split_path = file.split("/")
    if len(split_path) == 1:
        directory = "./"
    else:
        file = split_path[-1]
        directory = "/".join(split_path[:-1])
    del split_path
    file_format = file.split(".")[-1]
    if len(file_format) + 1 >= len(file):
        file_format = None
    if recursive:
        pathname = directory + "/**/" + file
    else:
        pathname = directory + "/" + file
    files = glob.glob(pathname, recursive=recursive)
    if len(files) > 1:
        print(
            "There are multiple files with '"
            + file
            + "' name in the directory/subdirectories"
        )  # TODO check multi extension
        raise Exception
    elif len(files) == 0:
        extract = pd.DataFrame()
        print("Warning : The file '" + file + "' wasn't found !")
    else:
        file = files[0].split(file)[0] + file
        if file_format in ("xlsx", "xls", "xlsm"):
            excel = pd.ExcelFile(file)
            sheets = excel.sheet_names
            try:
                if type(sheet) is int:
                    extract = excel.parse(sheets[sheet - 1], skiprows=skiprows)
                else:
                    extract = excel.parse(sheet, skiprows=skiprows)
            except IndexError:
                print(
                    "There is no sheet number "
                    + str(sheet)
                    + ", please select a valid sheet."
                )
                raise IndexError
            extract.dropna(how="all", inplace=True)
            if len(
                [
                    True
                    for elem in extract.columns
                    if (type(elem) is str and "Unnamed" in elem)
                ]
            ) == len(extract.columns):
                extract, extract.columns = (
                    extract.drop(extract.head(1).index),
                    extract.head(1).values.tolist()[0],
                )
            if classic_data:
                if on != "index" and on is not None:
                    extract.dropna(subset=[on], inplace=True)
                extract = caster(extract)
            extract.set_index(extract.columns[0])
            extract.drop(extract.columns[0], axis=1, inplace=True)
        elif file_format == "pkl":
            if not os.path.isfile(file):
                print("The specified pickle file doesn't exist !")
            extract = pd.read_pickle(file)
        elif file_format == "csv":
            from detect_delimiter import detect

            with open(file) as myfile:
                firstline = myfile.readline()
                delimiter = detect(firstline, default=";")
                myfile.close()
            extract = pd.read_csv(file, delimiter=delimiter)

            extract.dropna(how="all", inplace=True)
            if len(
                [
                    True
                    for elem in extract.columns
                    if (type(elem) is str and "Unnamed" in elem)
                ]
            ) == len(extract.columns):
                extract, extract.columns = (
                    extract.drop(extract.head(1).index),
                    extract.head(1).values.tolist()[0],
                )
            if classic_data:
                if on != "index":
                    extract.dropna(subset=[on], inplace=True)
                extract = caster(extract)
        elif file_format == "parquet":
            extract = pq.read_pandas(file).to_pandas()
        elif file_format in ["png", "jpg"]:
            from cv2 import imread

            img = imread(file)
            if (img[:, :, 0] == img[:, :, 1]).all() and (
                img[:, :, 0] == img[:, :, 2]
            ).all():
                return img[:, :, 0]
            return img
        elif file_format == "bmp":
            import imageio

            return imageio.v3.imread(file)
        elif file_format == "json":
            import json

            with open(file) as user_file:
                file_contents = user_file.read()
            return json.loads(file_contents)
        elif file_format == "txt":
            with open(file) as f:
                lines = f.readlines()
            return lines
        elif file_format == None:  # Directory
            return [
                load(elem, sheet, skiprows, on, classic_data, recursive)
                for elem in glob.glob(file + "/*")
            ]
        else:
            print("I don't handle this file format yet, came back in a decade.")
            raise Exception
        if classic_data:
            if on == "index":
                extract = extract[~extract.index.duplicated(keep="first")]
            elif (not (on in extract)) and (on != None):
                print(
                    "There is no column name '"
                    + on
                    + "' in the sheet "
                    + str(sheet)
                    + " of the file '"
                    + file
                    + "' so I can't load this file? Try to change the 'on' parameter"
                )
                raise Exception
            else:
                extract.drop_duplicates(subset=on, keep="last", inplace=True)
            extract.replace("Null", np.nan, inplace=True)
    return extract


def save(
    object_to_save: pd.DataFrame | np.ndarray | dict,
    file: str,
    sheet="Data",
    keep_index=True,
):  # FIXME column width when save excel
    file_format = file.split(".")[-1]
    if file_format == "xlsx":
        try:
            writer = pd.ExcelWriter(file)
        except PermissionError:
            print(
                "I can't access the file '" + file + "', the "
                "solution may be to close it and retry."
            )
            raise Exception
        try:
            object_to_save.to_excel(writer, sheet_name=sheet, index=keep_index)
        except IOError:
            object_to_save.to_excel(writer, sheet_name=sheet, index=True)
        # for column in object_to_save:
        #    column_length = max(object_to_save[column].astype(str).map(len).max(), len(column))
        #    col_idx = object_to_save.columns.get_loc(column)
        #    writer.sheets[sheet].set_column(col_idx, col_idx, column_length)
        writer.save()
    elif file_format == "pkl":
        pd.to_pickle(object_to_save, "./" + file)
    elif file_format == "parquet":
        caster(object_to_save)
        table = pa.Table.from_pandas(object_to_save, preserve_index=False)
        pq.write_table(table, file + ".parquet")
    elif file_format == "png":
        from PIL import Image

        im = Image.fromarray(object_to_save)
        im.save(file)
    elif file_format == "json":
        with open(file, "w") as outfile:
            outfile.write(object_to_save)
    else:
        print("I don't handle this file format yet, came back in a decade.")
        raise Exception


def merge(
    df: pd.DataFrame,
    file: str | pd.DataFrame,
    sheet=1,
    on="index",
    skiprows=None,
    how: str = "outer",
    save_file: bool = False,
):
    if not (type(df) is pd.DataFrame):
        print("The df parameter must be a DataFrame.")
        raise Exception
    if type(file) is pd.DataFrame:
        dataBase = file
    else:
        dataBase = load(
            file,
            sheet=sheet,
            skiprows=skiprows,
            on=on,
        )
    if not dataBase.empty:
        columns = list(dataBase.columns)
        if on != "index":
            try:
                columns.remove(on)
            except ValueError:
                print(
                    "You can't merge your data as there are not column '"
                    + on
                    + "' in your already loaded DataFrame."
                )
                raise Exception
        if df.empty:
            merge = dataBase.copy()
        else:
            merge = dataBase.merge(df, how=how, left_index=True, right_index=True)
        merge = merge.loc[:, ~merge.columns.duplicated()]
        col = list(merge.columns)
        if on != "index":
            col.remove(on)
        merge.dropna(how="all", subset=col, inplace=True)
        del dataBase
        drop_y = list(filter(lambda v: re.match(".*_y$", v), merge.columns))
        keep_x = list(filter(lambda v: re.match(".*_x$", v), merge.columns))
        keep = [name[:-2] for name in keep_x]
        for col_x, col_y in zip(keep_x, drop_y):
            merge[col_x] = merge[col_x].mask(pd.isnull, merge[col_y], errors="ignore")
        merge.drop(drop_y, axis=1, inplace=True)
        merge.rename(columns=dict(zip(keep_x, keep)), inplace=True)
        merge = merge.sort_index()
        merge.replace("Null", np.nan, inplace=True)
    else:
        merge = df
    if save_file:
        save(merge, file)
    else:
        return merge


def interpolator(df: pd.DataFrame) -> pd.DataFrame:
    array = np.array(df)
    filled_array = array[~np.isnan(array).any(axis=1), :]
    for i in range(array.shape[1]):
        idxs = list(range(array.shape[1]))
        idxs.pop(i)
        my_interpolator = NearestNDInterpolator(
            filled_array[:, idxs], filled_array[:, i], rescale=True
        )
        array[:, i] = np.apply_along_axis(
            lambda row: my_interpolator(*row[idxs]) if np.isnan(row[i]) else row[i],
            1,
            array,
        )
    return caster(pd.DataFrame(array, columns=df.columns, index=df.index))


def prepare_table(
    df: pd.DataFrame,
    y: str | list[str] = "",
    correlation_threshold: float = 0.9,
    missing_rows_threshold: float = 0.9,
    missing_columns_threshold: float = 0.6,
    categories_ratio_threshold: float = 0.1,
    id_correlation_threshold: float = 0.04,
    verbose: bool = False,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Prepare the table to feed a ML model.
    :param df: str, Optional
    :param y:
    :param correlation_threshold:
    :param missing_rows_threshold:
    :param missing_columns_threshold:
    :param categories_ratio_threshold:
    :param id_correlation_threshold:
    :return:
    """
    y = [y] if not y is list else y
    category_columns = df.select_dtypes(include=["category", object]).columns
    # Drop rows with NaNs in tye y columns
    if y != [""]:
        for col in y:
            df = df[df[col].notna()]
    # Drop the categorical columns with too much categories
    df = df.drop(
        [
            col
            for col in category_columns
            if df[col].nunique() / len(df) > categories_ratio_threshold
        ],
        axis=1,
    )
    # Convert categorical data to numerical ones
    df = pd.get_dummies(df)
    # Drop columns with not enough data
    df = df.loc[
        :,
        df.apply(
            lambda col: (1 - col.count() / len(df.index)) < missing_columns_threshold,
            axis=0,
        ),
    ]
    # Drop rows with not enough data
    df = df[
        df.apply(
            lambda row: (1 - row.count() / len(df.columns)) < missing_rows_threshold,
            axis=1,
        )
    ]
    correlate_couples = []
    corr = df.corr().abs()
    for col in corr:
        for index, val in corr[col].items():
            if val > correlation_threshold and (index != col):
                if (index, col) not in correlate_couples:
                    correlate_couples.append((col, index))
        if (df[col].nunique() == df.shape[0]) and (
            (corr[col].sum() - 1) / corr.shape[0] < id_correlation_threshold
        ):  # TODO améliorer
            # Drop ids columns (unique values with low correlations with other columns)
            if not col in y:
                df = df.drop(col, axis=1)
    for couple in correlate_couples:
        # Drop a column if highly correlated with another one
        if not any(elem in couple for elem in y):
            df = df.drop(couple[0], axis=1)
    if verbose:
        print("")
    return interpolator(df.drop(y, axis=1)), df[y]


def color(
    text: str,
    fg: Literal[
        "BLACK",
        "RED",
        "GREEN",
        "YELLOW",
        "BLUE",
        "MAGENTA",
        "CYAN",
        "WHITE",
        "LIGHTBLACK",
        "LIGHTRED",
        "LIGHTGREEN",
        "LIGHTYELLOW",
        "LIGHTBLUE",
        "LIGHTMAGENTA",
        "LIGHTCYAN",
        "LIGHTWHITE",
    ],
    bg: Literal[
        "BLACK",
        "RED",
        "GREEN",
        "YELLOW",
        "BLUE",
        "MAGENTA",
        "CYAN",
        "WHITE",
        "LIGHTBLACK",
        "LIGHTRED",
        "LIGHTGREEN",
        "LIGHTYELLOW",
        "LIGHTBLUE",
        "LIGHTMAGENTA",
        "LIGHTCYAN",
        "LIGHTWHITE",
    ] = None,
) -> str:
    if "LIGHT" in bg:
        bg += "_EX"
    if "LIGHT" in fg:
        fg += "_EX"
    bg = Back.__getattribute__(bg)
    colored_text = f"{Fore.__getattribute__(fg)}{text}{Style.RESET_ALL}"
    return colored_text if bg is None else bg + colored_text



from abc import abstractmethod
from logging.handlers import RotatingFileHandler
from os import remove
from threading import Thread
from time import sleep
from traceback import extract_stack, walk_stack
import logging
from typing import Literal

available_log_levels = Literal["PROD", "NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]


class Logger:
    """
    Abstract class to implement a configurable logger to a specific class
    The logger automatically creates a log file named like the class implementing this class

    :param debug: bool, default False

    :param log: str, default "PROD"
        Level of the logger among :
                PROD: Deactivate the logger
                NOTSET: Level 0
                DEBUG: Level 10
                INFO: Level 20
                WARNING: Level 30
                ERROR: Level 40
                CRITICAL: Level 50
    :param log_stdout: bool, default False
        Print the logs if enabled, however in both cases save the logs in the log file
    :param keep_log: bool, default False
        Overwrite the previous log file if False
    """
    @abstractmethod
    def _raiser(self, exception: str):
        raise NotImplementedError

    def __init__(self,
                 debug=False,
                 log: available_log_levels = "PROD",
                 log_stdout=False,
                 keep_log=False
                 ):
        self._DEBUG = debug
        self._LOG = log
        self._log_stdout = log_stdout
        self._keep_log = keep_log

        if self._LOG != "PROD":
            frame = None
            for frame, _ in walk_stack(None):
                variable_names = frame.f_code.co_varnames
                if variable_names == ():
                    break
                if frame.f_locals[variable_names[0]] not in (self, self.__class__):
                    break
                    # if the frame is inside a method of this instance,
                    # the first argument usually contains either the instance
                    # or its class, we want to find the first frame, where
                    # this is not the case
            else:
                self._raiser("No suitable outer frame found.")
            self._outer_frame = frame
            self.creation_module = frame.f_globals["__name__"]
            (
                self.creation_file,
                self.creation_line,
                self.creation_function,
                self.creation_text,
            ) = extract_stack(frame, 1)[0]
            self._creation_name = self.creation_text.split("=")[0].strip()
            super().__init__()
            Thread(target=self._check_existence_after_creation).start()
            try:
                if not keep_log:
                    remove(self._creation_name + ".log")
            except OSError:
                pass
            self._logger = logging.getLogger()
            self._logger.setLevel(logging.DEBUG)
            formatter = logging.Formatter(
                "%(asctime)s :: %(levelname)s :: %(message)s", "%d: %H:%M:%S"
            )
            file_handler = RotatingFileHandler(
                self._creation_name + ".log",
                "a",
                1000000,
                1,
                encoding="utf-8",
                )
            log_type = {
                "NOTSET": logging.NOTSET,
                "DEBUG": logging.DEBUG,
                "INFO": logging.INFO,
                "WARNING": logging.WARNING,
                "ERROR": logging.ERROR,
                "CRITICAL": logging.CRITICAL,
            }[self._LOG]
            file_handler.setLevel(log_type)
            file_handler.setFormatter(formatter)
            self._logger.addHandler(file_handler)
            # creation of a second handler which redirect every log of self.logging on the console
            if log_stdout:
                stream_handler = logging.StreamHandler()
                stream_handler.setLevel(log_type)
                self._logger.addHandler(stream_handler)
            self._log("Initialization of a new object of this instance", log="INFO")

    def _log(self, message, condition: bool = True, log="DEBUG"):
        """
        Create a log message
        :param message: str
            Content of the log message
        :param condition: bool, default True
            Condition to create the log
        :param log: str
            Log level (PROD, NOTSET, DEBUG, INFO, WARNING, ERROR, CRITICAL)
        :return: None
        """
        if condition and self._LOG != "PROD":
            logging_levels = {
                "NOTSET": logging.NOTSET,
                "DEBUG": logging.DEBUG,
                "INFO": logging.INFO,
                "WARNING": logging.WARNING,
                "ERROR": logging.ERROR,
                "CRITICAL": logging.CRITICAL,
            }[log]
            self._logger.log(logging_levels, message)
            logging.shutdown()

    def _check_existence_after_creation(self):
        """
        Check if the log file has been created, else raise an error
        :return: None
        """
        while self._outer_frame.f_lineno == self.creation_line:
            sleep(0.01)
        # this is executed as soon as the line number changes
        # now we can be sure the instance was actually created

        def error():
            self._raiser(
                "The creation name hasn't be found, the "
                "probable cause may be that you declared "
                "several variables at the same time."
            )

        nameparts = self._creation_name.split(".")
        var = None
        try:
            var = self._outer_frame.f_locals[nameparts[0]]
        except KeyError:
            error()
        finally:
            del self._outer_frame
        # make sure we have no permanent inter frame reference
        # which could hinder garbage collection
        try:
            for name in nameparts[1:]:
                var = getattr(var, name)
        except AttributeError:
            error()
        if var is not self:
            error()



from typing import Literal, Union

import pandas as pd
from pandas import DataFrame
from pylatex import Document, Section, Subsection, Figure, Package, NoEscape, SubFigure, Command, Table

import os

from ipoly.tracebackHandler import TracebackHandler

available_functions = Literal['section', 'subsection', 'image', 'text', 'table']


def nothing(*_args, **_kwargs):
    pass


class LaTeX(TracebackHandler):
    """
    Class for creating pdf files easily from Python by writing LaTeX code and compiling it.

    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, *kwargs)
        self.doc = None
        self.chain = []

    def _image(self, func, image_path: Union[str, tuple[str]] = None, caption: Union[str, tuple[str]] = None,
               centering: bool = True):
        if type(image_path) is str:
            image_path = tuple(image_path)
            caption = tuple(caption)
        elif caption is None:
            caption = tuple(None for _ in range(0, len(image_path)))
        image_filenames = tuple(os.path.join(os.path.dirname(__file__), "..\\" + path) for path in image_path)

        def image_func(*args, **kwargs):
            with self.doc.create(Figure(position='H')) as _figure:
                if centering:
                    self.doc.append(Command('centering'))
                for image_filename, image_caption in zip(image_filenames, caption):
                    with self.doc.create(SubFigure(width=NoEscape(rf'{1 / len(caption)}\linewidth'))) as image:
                        image.add_image(image_filename, width=NoEscape(r'0.95\linewidth'))
                        if image_caption:
                            image.add_caption(image_caption)
            func(*args, *kwargs)

        return image_func

    def _section(self, func, name: str):
        def section_func(*args, **kwargs):
            with self.doc.create(Section(name)):
                func(*args, *kwargs)

        return section_func

    def _subsection(self, func, name: str):
        def section_func(*args, **kwargs):
            with self.doc.create(Subsection(name)):
                func(*args, *kwargs)

        return section_func

    def _text(self, func, text: str):
        if text[-4:] == ".txt":
            with open(text, 'r') as file:
                text = file.read()

        def text_func(*args, **kwargs):
            self.doc.append(text)
            func(*args, *kwargs)

        return text_func

    def _table(self, func, table: Union[str, DataFrame], name: str = None):
        if type(table) is str:
            try:
                if table[-4:] == '.csv':
                    table = pd.read_csv(table)
                elif table[-4:] == '.pkl':
                    table = pd.read_pickle(table)
                elif table[-5:] == '.xlsx':
                    # noinspection PyArgumentList
                    table = pd.read_excel(table, index_col=0)

            except PermissionError:
                self._raiser(f"The file '{table}' is already open, please close it before trying again.")

        table = table.applymap(lambda x: "".join(c for c in x if c in
                                                 ['\t', '\n', '\r'] +
                                                 list(map(chr, range(224, 249))) +
                                                 list(map(chr, range(32, 35))) +
                                                 list(map(chr, range(38, 127)))) if type(x) == str else x)
        table = table.applymap(lambda x: x.replace("&", "\\&").replace("_", "\\_") if type(x) is str else x)

        def table_func(*args, **kwargs):
            with self.doc.create(Table(position='H')) as table_figure:
                if name:
                    table_figure.add_caption(name)
                table_figure.append(Command('centering'))
                table_figure.append(NoEscape(table.style.to_latex()))
            func(*args, *kwargs)

        return table_func

    def _chainer(self):
        self.doc = Document(inputenc='latin1')
        self.doc.packages.append(Package('geometry', options=['tmargin=1cm',
                                                              'lmargin=1cm',
                                                              'rmargin=1cm',
                                                              'bmargin=2cm']))
        self.doc.packages.append(Package('float'))
        self.doc.packages.append(Package('babel', options=['french']))
        self.doc.append(r'\listoffigures')
        func_chain = nothing
        for func, args, kwargs in reversed(self.chain):
            func_chain = func(func_chain, *args, *kwargs)
        func_chain(nothing)

    @TracebackHandler._traceback
    def add(self, func: available_functions, *args, mandatory: bool = True, **kwargs):
        funcs = {
            'section': self._section,
            'subsection': self._subsection,
            'image': self._image,
            'text': self._text,
            'table': self._table
        }
        if func == "image":
            if type(args[0]) == str:
                paths = tuple(args[0])
            else:
                paths = args[0]
            for path in paths:
                if not os.path.exists(os.path.join(os.path.dirname(__file__), "..\\" + path)):
                    if not mandatory:
                        return None
                    self._raiser(f"The path '{path}' doesn't correspond to an image.")
            if type(args[0]) == tuple and len(args) > 1 and len(args[0]) != len(args[1]):
                if not mandatory:
                    return None
                self._raiser("The number of images must be equal to the number of captions.")
        if func == "table":
            if not os.path.exists(os.path.join(os.path.dirname(__file__), "..\\" + args[0])):
                if not mandatory:
                    return None
                self._raiser(f"The provided path '{args[0]}' for the table is incorrect.")
            elif args[0][-4:] not in ('.csv', '.pkl', 'xlsx'):
                if not mandatory:
                    return None
                self._raiser(f"The extension of the file '{args[0]}' is not handled.")
        self.chain.append((funcs[func], args, kwargs))

    @TracebackHandler._traceback
    def generate_pdf(self, filepath: str = None, compiler="pdfLaTeX"):
        self._chainer()
        self.doc.generate_pdf(filepath, clean_tex=True, compiler=compiler)

    @TracebackHandler._traceback
    def generate_tex(self, filepath: str = None):
        self._chainer()
        self.doc.generate_tex(filepath)



import random
import os
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from pandas import DataFrame
from typing import Iterable, Any, Tuple
from sklearn import metrics
from nptyping import NDArray, Int
from plyer import notification


def set_seed(seed: int = 42) -> None:
    from tensorflow import random as tfr
    from torch.backends import cudnn
    from torch.cuda import manual_seed as cuda_manual_seed
    from torch import manual_seed
    from os import environ

    random.seed(seed)
    np.random.seed(seed)
    environ["PYTHONHASHSEED"] = str(seed)
    tfr.set_seed(seed)
    cuda_manual_seed(seed)
    manual_seed(seed)
    cudnn.deterministic = True
    cudnn.benchmark = False


def plot_history(history, keys: list[str] = None) -> None:
    if keys is None:
        keys = history.history.keys()
    for key in keys:
        plt.plot(history.history[key])
        plt.plot(history.history["val_accuracy"])
        plt.legend(["train", "val"], loc="upper left")
        plot("model accuracy", "epoch", "accuracy")


def plot_correlation_matrix(df: DataFrame) -> None:
    corr = df.corr()
    # Generate a mask for the upper triangle
    mask = np.triu(np.ones_like(corr, dtype=bool))
    # Set up the matplotlib figure
    f, ax = plt.subplots(figsize=(11, 9))
    # Generate a custom diverging colormap
    cmap = sns.diverging_palette(230, 20, as_cmap=True)
    # Draw the heatmap with the mask and correct aspect ratio
    sns.heatmap(
        corr,
        mask=mask,
        cmap=cmap,
        vmax=0.3,
        center=0,
        square=True,
        linewidths=0.5,
        cbar_kws={"shrink": 0.5},
    )


def _confusion_matrix(confusion_matrix, axes, class_names, fontsize=14):

    df_cm = DataFrame(
        confusion_matrix,
        index=class_names,
        columns=class_names,
    )

    try:
        heatmap = sns.heatmap(df_cm, annot=True, fmt="d", cbar=False, ax=axes)
    except ValueError:
        raise ValueError("Confusion matrix values must be integers.")
    heatmap.yaxis.set_ticklabels(
        heatmap.yaxis.get_ticklabels(), rotation=0, ha="right", fontsize=fontsize
    )
    heatmap.xaxis.set_ticklabels(
        heatmap.xaxis.get_ticklabels(), rotation=45, ha="right", fontsize=fontsize
    )
    axes.set_ylabel("True label")
    axes.set_xlabel("Predicted label")


def plot_confusion_matrix(
    y_pred: NDArray[Any, Int], y_true: NDArray[Any, Int], labels: Iterable[str]
):
    fig, ax = plt.subplots(2, 2, figsize=(10, 7))
    if ((np.unique(y_true.sum(axis=1)) == 1).all()) or (len(y_true.shape) == 1):
        if len(y_true.shape) == 2:
            y_true = y_true.argmax(axis=1)
            y_pred = y_pred.argmax(axis=1)
        _confusion_matrix(
            metrics.confusion_matrix(y_true, y_pred), ax.flatten(), ["N", "Y"]
        )
    else:
        confusion_matrix = metrics.multilabel_confusion_matrix(y_true, y_pred)
        for axes, cfs_matrix, label in zip(ax.flatten(), confusion_matrix, labels):
            _confusion_matrix(cfs_matrix, axes, ["N", "Y"])
            axes.set_title("Confusion Matrix for the class - " + label)

        fig.tight_layout()
        plt.show()


def croper(image: np.array, margin: int = 18):
    if len(np.unique(image)) == 1:
        raise Exception("The image is composed of a single color.")
    if len(image.shape) == 3:
        image_sum = image.sum(axis=2) % 765
    else:
        image_sum = image == 0
    true_points = np.argwhere(image_sum)
    top_left = true_points.min(axis=0)
    bottom_right = true_points.max(axis=0)
    return image[
        max(0, top_left[0] - margin) : bottom_right[0] + 1 + margin,
        max(0, top_left[1] - margin) : bottom_right[1] + 1 + margin,
    ]


def get_weighted_loss(pos_weights, neg_weights, epsilon=1e-7):
    """
    This function will calculate the loss function of the model
    Input :
        - pos_weights : positive frequencies wights
        - neg_weights : negative frequencies wights
        - epsilon : to not devide by 0
    Output :
        - loss : the loss classic function for the lost function.
    """
    from keras import backend as K
    from tensorflow import cast, float32

    def weighted_loss(y_true, y_pred):
        y_true, y_pred = cast(y_true, float32), cast(y_pred, float32)
        # initialize loss to zero
        loss = 0.0

        for i in range(len([pos_weights])):
            # for each class, add average weighted loss for that class
            loss += K.mean(
                -(
                    (pos_weights * y_true * K.log(y_pred + epsilon))
                    + (neg_weights * (1 - y_true) * K.log(1 - y_pred + epsilon))
                )
            )  # complete this line
        return loss

    return weighted_loss


def plot(
    title: str = None,
    xlabel: str = None,
    ylabel: str = None,
    figsize: Any = None,
    grid: bool = False,
    xticks: Tuple[int] = None,
    yticks: Tuple[int] = None,
):
    fig = plt.figure(figsize=figsize, dpi=100)
    fig.patch.set_facecolor("xkcd:white")
    plt.xticks(rotation=45, ha="right")

    plt.legend(bbox_to_anchor=(1.05, 1))
    if xticks:
        plt.yticks(np.arange(*xticks))
    if yticks:
        plt.yticks(np.arange(*yticks))
    plt.grid(grid)
    if ylabel:
        plt.ylabel(ylabel)
    if xlabel:
        plt.xlabel(xlabel)
    if title:
        plt.title(title)
    if ".png" in title:
        plt.savefig(title)
    else:
        plt.show()
    plt.close()


def say(message: str) -> None:
    from string import ascii_letters

    if all(
        char
        in ascii_letters
        + "0123456789,;:.?!-_ÂÃÄÀÁÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖØÙÚÛÜÝÞßàáâãäåæçèéêëìíîïðñòóôõöøùúûüýþÿ"
        for char in message
    ):
        raise Exception(
            "This message will not be said because you used some symbols that may be used for doing some malicious injection :("
        )
    from platform import system as ps
    from os import system as oss

    match ps():
        case "Windows":
            from win32com.client import Dispatch

            speak = Dispatch("SAPI.SpVoice").Speak
            speak(message)
        case "Darwin":
            oss(f"say '{message}' &")
        case "Linux":
            oss(f"spd-say '{message}' &")
        case syst:
            raise RuntimeError("Operating System '%s' is not supported" % syst)


def notify(message: str, title: str = "Hey!") -> None:
    notification.notify(
        app_name="iPoly",
        app_icon=os.path.join(os.path.dirname(__file__), r"img\ipoly.ico"),
        title=title,
        message=message,
        ticker="iPoly ticker",
        timeout=15,
    )


def path(path: str) -> str:
    from platform import system as ps

    if ps() == "Windows":
        return path.replace("/", "\\")
    return path.replace("\\", "/")



from typeguard.importhook import install_import_hook

with install_import_hook("ipoly"):
    from ipoly.file_management import *
    from ipoly.LaTeX import *
    from ipoly.email import *
    from ipoly.logger import *
    from ipoly.tracebackHandler import *
    from ipoly.ml import *



from sys import exc_info
from typing import Any

from IPython import InteractiveShell

from ipoly.logger import Logger


class Error(Exception):
    """Base class for exceptions in this module."""
    pass


class TracebackHandler(Logger):
    """
    Abstract class aiming to deactivate the Traceback if the error is deliberately raised with the '_raiser' method.
    This class implements the Logger class.
    The Traceback is restored if an error is not raised by the '_raiser' method.
    Each public method of the class implementing 'TracebackHandler' should be decorated by the '_traceback' method to
    take advantage of its features.

    :param debug: bool, default False
    :param kwargs:
    """
    def __init__(self, debug=False, **kwargs):
        _trace = None
        _ipython = None
        _hide_traceback = None
        self._debug = debug
        if not self._debug:
            self._ipython = self._get_ipython()

            def hide_traceback(**_kwargs):
                error_type, value, tb = exc_info()
                value.__cause__ = None  # suppress chained exceptions
                return self._ipython._showtraceback(
                    error_type,
                    value,
                    self._ipython.InteractiveTB.get_exception_only(error_type, value),
                )

            self.hide_traceback = hide_traceback
            if self._ipython is not None:
                self._trace = self._ipython.showtraceback

        super().__init__(**kwargs)

    def _raiser(self, exception: str) -> None:
        self._log(exception, log="CRITICAL")
        if not self._debug and (self._ipython is not None):
            self._ipython.showtraceback = self.hide_traceback
        raise Error(exception)

    @staticmethod
    def _traceback(func) -> Any:
        """
        Decorator method to deactivate temporarily the Traceback and restore it at the end of the decorated method.
        """
        def wrapper(self, *args, **kwargs):
            if not self._DEBUG and self._ipython is not None:
                self._ipython.showtraceback = self._trace
            try:
                return func(self, *args, **kwargs)
            except Error as e:
                raise e
            except BaseException as e:
                self._log(e, log="CRITICAL")
                raise e

        return wrapper

    @staticmethod
    def _get_ipython() -> InteractiveShell:
        from IPython import get_ipython
        return get_ipython()
